@@ -10,7 +10,7 @@
  * See the file "license.terms" for information on usage and redistribution
  * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
  *
- * RCS: @(#) $Id: tclUnixFCmd.c,v 1.28.2.8 2005/12/05 15:10:33 dgp Exp $
+ * RCS: @(#) $Id: tclUnixFCmd.c,v 1.28.2.13 2006/09/07 18:49:28 vasiljevic Exp $
  *
  * Portions of this code were derived from NetBSD source code which has
  * the following copyright notice:
@@ -189,6 +189,22 @@ Realpath(path, resolved)
 #define Realpath realpath
 #endif
 
+#ifndef NO_REALPATH
+#if defined(__APPLE__) && defined(TCL_THREADS) && \
+	defined(MAC_OS_X_VERSION_MIN_REQUIRED) && \
+	MAC_OS_X_VERSION_MIN_REQUIRED < 1030
+/*
+ * prior to Darwin 7, realpath is not threadsafe, c.f. bug 711232;
+ * if we might potentially be running on pre-10.3 OSX,
+ * check Darwin release at runtime before using realpath.
+ */
+extern long tclMacOSXDarwinRelease;
+#define haveRealpath (tclMacOSXDarwinRelease >= 7)
+#else
+#define haveRealpath 1
+#endif
+#endif /* NO_REALPATH */
+
 
 /*
  *---------------------------------------------------------------------------
@@ -266,7 +282,7 @@ DoRenameFile(src, dst)
      * conditionally compiled because realpath() not defined on all systems.
      */
 
-    if (errno == EINVAL) {
+    if (errno == EINVAL && haveRealpath) {
 	char srcPath[MAXPATHLEN], dstPath[MAXPATHLEN];
 	DIR *dirPtr;
 	Tcl_DirEntry *dirEntPtr;
@@ -402,6 +418,9 @@ DoCopyFile(src, dst, statBufPtr)
 		return TCL_ERROR;
 	    }
 #ifdef HAVE_COPYFILE
+#ifdef WEAK_IMPORT_COPYFILE
+	    if (copyfile != NULL)
+#endif
 	    copyfile(src, dst, NULL, COPYFILE_XATTR|COPYFILE_NOFOLLOW_SRC);
 #endif
 	    break;
@@ -1206,6 +1225,9 @@ CopyFileAtts(src, dst, statBufPtr)
 	return TCL_ERROR;
     }
 #ifdef HAVE_COPYFILE
+#ifdef WEAK_IMPORT_COPYFILE
+    if (copyfile != NULL)
+#endif
     copyfile(src, dst, NULL, COPYFILE_XATTR|COPYFILE_ACL);
 #endif
     return TCL_OK;
@@ -1249,8 +1271,9 @@ GetGroupAttribute(interp, objIndex, fileName, attributePtrPtr)
 	return TCL_ERROR;
     }
 
-    groupPtr = getgrgid(statBuf.st_gid);		/* INTL: Native. */
-    if (groupPtr == NULL) {
+    groupPtr = TclpGetGrGid(statBuf.st_gid);
+
+    if (result == -1 || groupPtr == NULL) {
 	*attributePtrPtr = Tcl_NewIntObj((int) statBuf.st_gid);
     } else {
 	Tcl_DString ds;
@@ -1301,8 +1324,9 @@ GetOwnerAttribute(interp, objIndex, fileName, attributePtrPtr)
 	return TCL_ERROR;
     }
 
-    pwPtr = getpwuid(statBuf.st_uid);			/* INTL: Native. */
-    if (pwPtr == NULL) {
+    pwPtr = TclpGetPwUid(statBuf.st_uid);
+
+    if (result == -1 || pwPtr == NULL) {
 	*attributePtrPtr = Tcl_NewIntObj((int) statBuf.st_uid);
     } else {
 	Tcl_DString ds;
@@ -1394,9 +1418,8 @@ SetGroupAttribute(interp, objIndex, fileName, attributePtr)
 	int length;
 
 	string = Tcl_GetStringFromObj(attributePtr, &length);
-
 	native = Tcl_UtfToExternalDString(NULL, string, length, &ds);
-	groupPtr = getgrnam(native);			/* INTL: Native. */
+	groupPtr = TclpGetGrNam(native); /* INTL: Native. */
 	Tcl_DStringFree(&ds);
 
 	if (groupPtr == NULL) {
@@ -1457,12 +1480,12 @@ SetOwnerAttribute(interp, objIndex, fileName, attributePtr)
 	int length;
 
 	string = Tcl_GetStringFromObj(attributePtr, &length);
-
 	native = Tcl_UtfToExternalDString(NULL, string, length, &ds);
-	pwPtr = getpwnam(native);			/* INTL: Native. */
+	pwPtr = TclpGetPwNam(native); /* INTL: Native. */
 	Tcl_DStringFree(&ds);
 
 	if (pwPtr == NULL) {
+	    endpwent();
 	    Tcl_AppendResult(interp, "could not set owner for file \"",
 			     Tcl_GetString(fileName), "\": user \"", 
 			     string, "\" does not exist",
@@ -1474,7 +1497,8 @@ SetOwnerAttribute(interp, objIndex, fileName, attributePtr)
 
     native = Tcl_FSGetNativePath(fileName);
     result = chown(native, (uid_t) uid, (gid_t) -1);   /* INTL: Native. */
-
+    
+    endpwent();
     if (result != 0) {
 	Tcl_AppendResult(interp, "could not set owner for file \"", 
 			 Tcl_GetString(fileName), "\": ", 
@@ -1815,15 +1839,23 @@ TclpObjNormalizePath(interp, pathPtr, nextCheckpoint)
 
 #ifndef NO_REALPATH
     /* For speed, try to get the entire path in one go */
-    if (nextCheckpoint == 0) {
+    if (nextCheckpoint == 0 && haveRealpath) {
         char *lastDir = strrchr(currentPathEndPosition, '/');
 	if (lastDir != NULL) {
 	    nativePath = Tcl_UtfToExternalDString(NULL, path, 
 						  lastDir - path, &ds);
 	    if (Realpath(nativePath, normPath) != NULL) {
-		nextCheckpoint = lastDir - path;
-		goto wholeStringOk;
+		if (*nativePath != '/' && *normPath == '/') {
+		    /*
+		     * realpath has transformed a relative path into an
+		     * absolute path, we do not know how to handle this.
+		     */
+		} else {
+		    nextCheckpoint = lastDir - path;
+		    goto wholeStringOk;
+		}
 	    }
+	    Tcl_DStringFree(&ds);
 	}
     }
     /* Else do it the slow way */
@@ -1861,61 +1893,60 @@ TclpObjNormalizePath(interp, pathPtr, nextCheckpoint)
      * have 'realpath'.
      */
 #ifndef NO_REALPATH
-    /* 
-     * If we only had '/foo' or '/' then we never increment nextCheckpoint
-     * and we don't need or want to go through 'Realpath'.  Also, on some
-     * platforms, passing an empty string to 'Realpath' will give us the
-     * normalized pwd, which is not what we want at all!
-     */
-    if (nextCheckpoint == 0) return 0;
-    
-    nativePath = Tcl_UtfToExternalDString(NULL, path, nextCheckpoint, &ds);
-    if (Realpath(nativePath, normPath) != NULL) {
-	int newNormLen;
-	wholeStringOk:
-	newNormLen = strlen(normPath);
-	if ((newNormLen == Tcl_DStringLength(&ds))
-		&& (strcmp(normPath, nativePath) == 0)) {
-	    /* String is unchanged */
-	    Tcl_DStringFree(&ds);
-	    if (path[nextCheckpoint] != '\0') {
-		nextCheckpoint++;
-	    }
-	    return nextCheckpoint;
-	}
+    if (haveRealpath) {
+	/* 
+	 * If we only had '/foo' or '/' then we never increment nextCheckpoint
+	 * and we don't need or want to go through 'Realpath'.  Also, on some
+	 * platforms, passing an empty string to 'Realpath' will give us the
+	 * normalized pwd, which is not what we want at all!
+	 */
+	if (nextCheckpoint == 0) return 0;
 	
-	/* 
-	 * Free up the native path and put in its place the
-	 * converted, normalized path.
-	 */
-	Tcl_DStringFree(&ds);
-	Tcl_ExternalToUtfDString(NULL, normPath, (int) newNormLen, &ds);
-
-	if (path[nextCheckpoint] != '\0') {
-	    /* not at end, append remaining path */
-	    int normLen = Tcl_DStringLength(&ds);
-	    Tcl_DStringAppend(&ds, path + nextCheckpoint,
-		    pathLen - nextCheckpoint);
+	nativePath = Tcl_UtfToExternalDString(NULL, path, nextCheckpoint, &ds);
+	if (Realpath(nativePath, normPath) != NULL) {
+	    int newNormLen;
+	    wholeStringOk:
+	    newNormLen = strlen(normPath);
+	    if ((newNormLen == Tcl_DStringLength(&ds))
+		    && (strcmp(normPath, nativePath) == 0)) {
+		/* String is unchanged */
+		Tcl_DStringFree(&ds);
+		if (path[nextCheckpoint] != '\0') {
+		    nextCheckpoint++;
+		}
+		return nextCheckpoint;
+	    }
+	    
 	    /* 
-	     * We recognise up to and including the directory
-	     * separator.
-	     */	
-	    nextCheckpoint = normLen + 1;
-	} else {
-	    /* We recognise the whole string */ 
-	    nextCheckpoint = Tcl_DStringLength(&ds);
+	     * Free up the native path and put in its place the
+	     * converted, normalized path.
+	     */
+	    Tcl_DStringFree(&ds);
+	    Tcl_ExternalToUtfDString(NULL, normPath, (int) newNormLen, &ds);
+    
+	    if (path[nextCheckpoint] != '\0') {
+		/* not at end, append remaining path */
+		int normLen = Tcl_DStringLength(&ds);
+		Tcl_DStringAppend(&ds, path + nextCheckpoint,
+			pathLen - nextCheckpoint);
+		/* 
+		 * We recognise up to and including the directory
+		 * separator.
+		 */	
+		nextCheckpoint = normLen + 1;
+	    } else {
+		/* We recognise the whole string */ 
+		nextCheckpoint = Tcl_DStringLength(&ds);
+	    }
+	    /* 
+	     * Overwrite with the normalized path.
+	     */
+	    Tcl_SetStringObj(pathPtr, Tcl_DStringValue(&ds),
+		    Tcl_DStringLength(&ds));
 	}
-	/* 
-	 * Overwrite with the normalized path.
-	 */
-	Tcl_SetStringObj(pathPtr, Tcl_DStringValue(&ds),
-		Tcl_DStringLength(&ds));
+	Tcl_DStringFree(&ds);
     }
-    Tcl_DStringFree(&ds);
 #endif	/* !NO_REALPATH */
 
     return nextCheckpoint;
 }
-
-
-
